#lang scheme

(define (comp expr)
  (cond
    [(number? expr) (lambda (env) expr)]
    [(symbol? expr)
     (lambda (env)
       (cdr (assq expr env)))]
    [else
     (case (car expr)
       [(+)
        (let ([a (comp (cadr expr))]
              [b (comp (caddr expr))])
          (lambda (env) (+ (a env) (b env))))]
       [(-)
        (let ([a (comp (cadr expr))]
              [b (comp (caddr expr))])
          (lambda (env) (- (a env) (b env))))]
       [(zero?)
        (let ([a (comp (cadr expr))])
          (lambda (env) (zero? (a env))))]
       [(if)
        (let ([a (comp (cadr expr))]
              [b (comp (caddr expr))]
              [c (comp (cadddr expr))])
          (lambda (env)
            (if (a env)
                (b env)
                (c env))))]
       [(let)
        (let ([rhs (comp (cadr (caadr expr)))]
              [body (comp (caddr expr))])
          (lambda (env)
            (let ([rhs-val (rhs env)])
              (body
               (cons (cons (car (caadr expr))
                           rhs-val)
                     env)))))]
       [(lambda)
        (let ([body (comp (caddr expr))])
          (lambda (env)
            (lambda (a)
              (body (cons (cons (caadr expr) a)
                          env)))))]
       [else
        (let ([a (comp (car expr))]
              [b (comp (cadr expr))])
          (lambda (env)
            (let ([clos (a env)]
                  [arg-val (b env)])
              (clos arg-val))))])]))

((comp '(let ([f (lambda (x) 
                   (lambda (y)
                     (- y x)))]) 
          (+ ((f 10) 2) 3)))
 '())

(time ((comp '(let ([Y (lambda (m)
                         ((lambda (f) (m (lambda (a) ((f f) a))))
                          (lambda (f) (m (lambda (a) ((f f) a))))))])
                (let ([sum
                       (Y (lambda (sum)
                            (lambda (n)
                              (if (zero? n)
                                  0
                                  (+ n (sum (- n 1)))))))])
                  (sum 50000))))
       '()))
